# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
# Copyright © 2020 Necdet Can Ateşman, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in
# the file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district
# the Licensee has his registered seat, an establishment or assets.


"""
This module handles :term:`templates <template>`, i.e. files that need to be
pre-processed before they can be put to their actual use. The module
initialization yields a :class:`.Renderer` object, which is the central hub
for rendering templates.
"""

from ._init import init, ConfiguredTplModule, FileType
from ._exc import TemplateNotFound
from .renderer import Renderer
from .loader import Loader, FileSystemLoader, ChainLoader, PrefixedLoader

__all__ = (
    'init', 'ConfiguredTplModule', 'FileType', 'TemplateNotFound', 'Renderer',
    'Loader', 'FileSystemLoader', 'ChainLoader', 'PrefixedLoader')



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
# Copyright © 2020 Necdet Can Ateşman, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in
# the file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district
# the Licensee has his registered seat, an establishment or assets.

import os
from ._exc import TemplateNotFound
from .loader import FileSystemLoader
from collections import namedtuple, defaultdict, OrderedDict
from score.init import (
    parse_list, extract_conf, ConfiguredModule, ConfigurationError)
import re


defaults = {
    'rootdirs': [],
}


def init(confdict):
    """
    Initializes this module acoording to :ref:`our module initialization
    guidelines <module_initialization>` with the following configuration keys:

    :confkey:`rootdirs` :confdefault:`None`
        Denotes the root folder containing all templates.
    """
    conf = dict(defaults.items())
    conf['rootdirs'] = []
    conf.update(confdict)
    if 'rootdir' in conf and conf['rootdirs']:
        import score.tpl
        raise ConfigurationError(
            score.tpl, "Both 'rootdir' and 'rootdirs' given")
    if 'rootdir' in conf:
        conf['rootdirs'] = conf['rootdir']
    rootdirs = parse_list(conf['rootdirs'])
    for rootdir in rootdirs:
        if not os.path.isdir(rootdir):
            import score.tpl
            raise ConfigurationError(
                score.tpl, 'Given rootdir is not a folder: %s' % (rootdir,))
    tpl = ConfiguredTplModule(rootdirs)
    extensions = set()
    for key in extract_conf(conf, 'filetype.'):
        extensions.add(key.rsplit('.', 1)[0])
    for ext in extensions:
        mimetype = conf.get('filetype.%s.mimetype' % ext)
        if mimetype is None:
            import score.tpl
            raise ConfigurationError(
                score.tpl, 'No mimetype configured for extension %s' % (ext,))
        tpl.filetypes[mimetype].extensions.append(ext)
    return tpl


class ConfiguredTplModule(ConfiguredModule):
    """
    This module's :class:`configuration class
    <score.init.ConfiguredModule>`.
    """

    def __init__(self, rootdirs):
        super().__init__(__package__)
        self.rootdirs = rootdirs
        self.filetypes = FileTypes(self)
        self.loaders = Loaders(self)
        self.engines = Engines(self)
        self._renderers = defaultdict(dict)

    def define_global(self, mimetype, name, value, escape=True):
        self.filetypes[mimetype].add_global(name, value, escape=escape)

    def iter_paths(self, mimetype=None):
        """
        Provides a generator iterating over all known template paths. If the
        optional parameter *mimetype* is present, only templates of that mime
        type will pe provided instead.
        """
        def all_paths():
            for mimetype in self.filetypes:
                for extension in self.filetypes[mimetype].extensions:
                    for loader in self.loaders[extension]:
                        yield from loader.iter_paths()

        def valid_paths():
            for path in all_paths():
                try:
                    self._find_filetype(path)
                    yield path
                except TemplateNotFound:
                    pass

        if mimetype:
            extensions = self.filetypes[mimetype].extensions
            regex = re.compile(r'(^|/)[^/.]+\.(%s)($|\.)' %
                               '|'.join(map(re.escape, extensions)))
            for path in valid_paths():
                if regex.search(path):
                    yield path
        else:
            yield from valid_paths()

    def load(self, path):
        """
        Loads given template *path*.

        See :meth:`Loader.load`.
        """
        return self._find_loader(path).load(path)

    def render(self, path, variables=None, *, apply_postprocessors=True):
        """
        Renders give template *path* with the optional `dict` of *variables*.
        It is possible to prevent running the file type's postprocessors by
        passing a falsy value for the *apply_postprocessors* parameter.
        """
        filetype = self._find_filetype(path)
        is_file, result = self.load(path)
        if variables is None:
            variables = {}
        for renderer in self._find_renderers(path, filetype=filetype):
            if is_file:
                result = renderer.render_file(result, variables, path=path)
                is_file = False
            else:
                result = renderer.render_string(result, variables, path=path)
        if is_file:
            result = open(result).read()
        if apply_postprocessors:
            for postprocessor in filetype.postprocessors:
                result = postprocessor(result)
        return result

    def mimetype(self, path):
        """
        Provides to mime type associated with given *path*.

        See :ref:`tpl_file_types`.
        """
        return self._find_filetype(path).mimetype

    def hash(self, path):
        """
        Provides a hash for given template *path*.

        See :meth:`Loader.hash`.
        """
        return self._find_loader(path).hash(path)

    def _finalize(self):
        # make sure that every file extension is associated with
        # at most one filetype
        all_extensions = set()
        for filetype in self.filetypes.values():
            duplicates = all_extensions.intersection(filetype.extensions)
            if not duplicates:
                all_extensions |= set(filetype.extensions)
                continue
            extension = duplicates.pop()
            filetypes = [f for f in self.filetypes
                         if extension in f.extensions.extensions]
            import score.tpl
            raise ConfigurationError(
                score.tpl,
                'Extension %s registered with multiple mimetypes: %s' % (
                    extension, filetypes))
        for extension in self.loaders:
            if extension not in all_extensions:
                import score.tpl
                raise ConfigurationError(
                    score.tpl,
                    'Loader Extension "%s" has no filetype' % (extension,))
        for extension in self.engines:
            if extension not in all_extensions:
                import score.tpl
                raise ConfigurationError(
                    score.tpl,
                    'Engine Extension "%s" has no filetype' % (extension,))
        # sort loaders and engines by length of extension string
        # this is important, as we want to test 'tar.gz' before 'gz'
        self.loaders = OrderedDict(
            (ext, self.loaders[ext])
            for ext in sorted(all_extensions, key=len, reverse=True))
        self.engines = OrderedDict(
            (ext, self.engines[ext])
            for ext in sorted(self.engines, key=len, reverse=True))

    def _find_loader(self, path):
        parts = os.path.basename(path).split('.', maxsplit=1)
        if len(parts) == 1:
            # TODO: other exception?
            raise TemplateNotFound(path)
        if parts[1] not in self.loaders:
            while '.' in parts[1]:
                more_parts = parts[1].split('.', maxsplit=1)
                parts[0] += '.' + more_parts[0]
                parts[1] = more_parts[1]
                if parts[1] in self.loaders:
                    break
            else:
                raise TemplateNotFound(path)
        for loader in self.loaders[parts[1]]:
            if loader.is_valid(path):
                return loader
        raise TemplateNotFound(path)

    def _find_renderers(self, path, *, filetype=None):
        renderers = []
        filename = os.path.basename(path)
        if filetype is None:
            filetype = self._find_filetype(path)
        while True:
            candidates = []
            for extension, engine in self.engines.items():
                idx = filename.rfind('.%s' % extension)
                if idx < 0:
                    continue
                candidates.append((extension, idx, engine))
            if not candidates:
                break
            extension, idx, engine = sorted(
                candidates, key=lambda x: (x[1] + len(x[0]), len(x[0])),
                reverse=True)[0]
            filename = filename[:len(extension) + 1]
            if filetype not in self._renderers[engine]:
                self._renderers[engine][filetype] = engine(self, filetype)
            renderers.append(self._renderers[engine][filetype])
        return renderers

    def _find_filetype(self, path):
        filename = os.path.basename(path)
        extensions = filename.split('.')[1:]
        for i in range(len(extensions), 0, -1):
            extension = '.'.join(extensions[:i])
            try:
                return next(filetype
                            for filetype in self.filetypes.values()
                            if extension in filetype.extensions)
            except StopIteration:
                pass
        raise TemplateNotFound(path)


class Loaders(defaultdict):

    def __init__(self, conf):
        self.conf = conf
        super().__init__()

    def __missing__(self, key):
        if '/' in key:
            raise ValueError('Invalid File extension "%s"' % (key,))
        loaders = list()
        if self.conf.rootdirs:
            loaders.append(FileSystemLoader(self.conf.rootdirs, key))
        self[key] = loaders
        return loaders


class Engines(dict):

    def __init__(self, conf):
        self.conf = conf
        super().__init__()

    def __setitem__(self, key, value):
        if key in self:
            raise ValueError(
                'Renderer for extension `%s` already defined as %s' %
                (key, repr(self[key])))
        if '/' in key:
            raise ValueError('Invalid File extension "%s"' % (key,))
        return super().__setitem__(key, value)


class FileTypes(defaultdict):

    def __init__(self, conf):
        self.conf = conf
        super().__init__()

    def __missing__(self, key):
        filetype = FileType(self.conf, key)
        self[key] = filetype
        return filetype


class FileType:
    """
    Represents a known :term:`file type`. Attributes may only be modified until
    :ref:`finalization <finalization>` of the module.
    """

    def __init__(self, conf, mimetype):
        self.__conf = conf
        self.__mimetype = mimetype
        self.__extensions = []
        self.__postprocessors = []
        self.__globals = []
        self.__finalized = False
        self.__escape = None

    def _finalize(self):
        # TODO: check for duplicates in extensions
        self.__extensions = tuple(self.__extensions)
        self.__postprocessors = tuple(self.__postprocessors)
        self.__globals = tuple(self.__globals)
        self.__finalized = True

    @property
    def mimetype(self):
        return self.__mimetype

    @property
    def extensions(self):
        return self.__extensions

    @extensions.setter
    def extensions(self, value):
        assert not self.__finalized
        self.__extensions = value

    @property
    def postprocessors(self):
        return self.__postprocessors

    @postprocessors.setter
    def postprocessors(self, value):
        assert not self.__finalized
        self.__postprocessors = value

    @property
    def escape(self):
        return self.__escape

    @escape.setter
    def escape(self, callback):
        assert not self.__finalized
        self.__escape = callback

    @property
    def globals(self):
        return self.__globals

    def add_global(self, name, value, *, escape=True):
        """
        Defines a new global variable with given *name* and *value* for this
        file type. The optional parameter *escape* determines, whether the
        value must be escaped in the final output.
        """
        assert not self.__finalized
        assert not any(x for x in self.__globals if x.name == name)
        self.__globals.append(VariableDefinition(name, value, escape))


VariableDefinition = namedtuple('VariableDefinition',
                                ('name', 'value', 'escape'))



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
# Copyright © 2020 Necdet Can Ateşman, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in
# the file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district
# the Licensee has his registered seat, an establishment or assets.

from ._exc import TemplateNotFound
import abc
import os
import xxhash


class Loader:
    """
    Object capable of loading template content.
    """

    @property
    def paths(self):
        return list(self.iter_paths())

    @abc.abstractmethod
    def iter_paths(self):
        """
        Provide a generator iterating all valid paths of this loader.
        """
        pass

    @abc.abstractmethod
    def load(self, path):
        """
        Load given *path*. Returns a 2-tuple, where the first value is a `bool`
        indicating whether the other value is a file path (`True`) or the
        contents of the template (`False`). Illustration of the return values:

        .. code-block:: python

            (True, '/path/to/the/template.html')
            (False, '<html>The contents of the template</html>')

        Raises :class:`TemplateNotFound` if this loader cannot load the given
        path.
        """
        pass

    def is_valid(self, path):
        """
        Whether given *path* is valid for this loader, i.e. whether a call to
        :meth:`load` would return successfully.
        """
        return path in self.iter_paths()

    def hash(self, path):
        """
        Provides a random `str`, that will always change whenever the file
        content changes. The default implementation uses the content's xxHash_.

        .. _xxHash: http://cyan4973.github.io/xxHash/
        """
        is_file, result = self.load(path)
        if is_file:
            try:
                hash = xxhash.xxh64()
                with open(result, 'rb') as file:
                    for chunk in iter(lambda: file.read(4096), b""):
                        hash.update(chunk)
                return hash.hexdigest()
            except FileNotFoundError:
                raise TemplateNotFound(path)
        else:
            if isinstance(result, str):
                result = result.encode('ASCII')
            return xxhash.xxh64(result).hexdigest()


class FileSystemLoader(Loader):
    """
    :class:`Loader` searching for files with a given *extension* inside given
    folders.
    """

    def __init__(self, rootdirs, extension):
        if isinstance(rootdirs, str):
            rootdirs = [rootdirs]
        self.rootdirs = rootdirs
        self.extension = extension

    def iter_paths(self):
        if not self.rootdirs:
            return
        found = []
        ext = '.%s' % (self.extension,)
        extlen = len(ext)
        for rootdir in self.rootdirs:
            for base, dirs, files in os.walk(rootdir, followlinks=True):
                base = os.path.relpath(base, rootdir)
                if base == '.':
                    base = ''
                else:
                    base = base + '/'
                for filename in files:
                    if filename[-extlen:] != ext:
                        continue
                    path = base + filename
                    if path in found:
                        continue
                    found.append(path)
                    yield path

    def load(self, path):
        for rootdir in self.rootdirs:
            fullpath = os.path.join(rootdir, path.lstrip('/'))
            relpath = os.path.relpath(fullpath, rootdir)
            if relpath.startswith(os.pardir + os.sep):
                # outside of rootdir
                continue
            if os.path.exists(fullpath):
                return True, fullpath
        raise TemplateNotFound(path)


class ChainLoader(Loader):
    """
    A :class:`Loader` that will wrap the other given *loaders* and simulate a
    loader, that combines the features of all of them. If this receives a
    Loader capable of loading 'a.tpl' and another Loader that can load 'b.tpl',
    this ChainLoader instance will be able to load 'a.tpl' *and* 'b.tpl'.
    """

    def __init__(self, loaders):
        self.loaders = loaders

    def is_valid(self, path):
        return any(loader.is_valid(path) for loader in self.loaders)

    def iter_paths(self):
        for loader in self.loaders:
            yield from loader.iter_paths()

    def load(self, path):
        for loader in self.loaders:
            try:
                return loader.load(path)
            except TemplateNotFound:
                pass
        raise TemplateNotFound(path)

    def hash(self, path):
        for loader in self.loaders:
            try:
                return loader.hash(path)
            except TemplateNotFound:
                pass


class PrefixedLoader(Loader):
    """
    A :class:`Loader` wrapper, that prefixes a pre-given path to all templates.
    If this is configured with the prefix 'foo/' and given a loader capable of
    loading 'a.tpl', the resulting instance will only be able to load
    'foo/a.tpl' (but not 'a.tpl').

    Note that this loader does not modify the given prefix. It is important to
    add a slash to the prefix, if you want it to behave like a folder, for
    example.
    """

    def __init__(self, prefix, wrapped):
        self.prefix = prefix
        self.wrapped = wrapped

    def iter_paths(self):
        yield from (self.prefix + path for path in self.wrapped.iter_paths())

    def load(self, path):
        if not path.startswith(self.prefix):
            raise TemplateNotFound(path)
        return self.wrapped.load(path[len(self.prefix):])

    def is_valid(self, path):
        if not path.startswith(self.prefix):
            return False
        return self.wrapped.is_valid(path[len(self.prefix):])

    def hash(self, path):
        if not path.startswith(self.prefix):
            raise TemplateNotFound(path)
        return self.wrapped.hash(path[len(self.prefix):])



# Copyright © 2015-2018 STRG.AT GmbH, Vienna, Austria
# Copyright © 2020 Necdet Can Ateşman, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in
# the file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district
# the Licensee has his registered seat, an establishment or assets.

from abc import ABCMeta, abstractmethod


class Renderer(metaclass=ABCMeta):

    def __init__(self, tpl_conf, filetype):
        self._tpl_conf = tpl_conf
        self.filetype = filetype

    def render_file(self, file, variables, path=None):
        """
        Renders given template *file* with the given *variables* dict.
        """
        return self.render_string(open(file).read(), variables, path=path)

    @abstractmethod
    def render_string(self, string, variables, path=None):
        """
        Renders the given template content *string* with the given *variables*
        dict.
        """
        return
